# 2

# 人工智能的基本用例

在本章中，我们将讨论**人工智能**（ **AI** ）的一些用例。 这绝不是详尽的清单。 许多行业已受到 AI 的影响，但尚未受影响的那些行业的清单每天都在缩短。 具有讽刺意味的是，机器人，自动化和 AI 无法接手的某些工作是薪水较低，需要较少“大脑”力量的工作。 例如，在我们能够代替发型师和水管工之前还需要一段时间。 这两项工作都需要机器人还需要掌握很多技巧和细节。 我知道，我的妻子要花很长时间才能将她的头发信任除当前发型师以外的任何人，更不用说机器人了。

本章将讨论：

*   一些具有代表性的 AI 用例
*   最长的工作将由自动化取代
*   受 AI 影响最大的行业

# 代表性 AI 用例

从金融到医学，很难找到一个不受人工智能干扰的行业。 我们将重点关注日常生活中最流行的 AI 应用的真实示例。 我们将探索当前的技术水平以及即将发生的事情。 最重要的是，也许这本书会激发您的想象力，并且您会提出一些对社会产生积极影响的新想法，我们可以将其添加到本书的下一版中。

人工智能，认知计算，机器学习和深度学习只是当今促成快速变化的一些破坏性技术。 由于云计算，**物联网**（ **IoT** ）和边缘计算的进步，可以更快地采用这些技术。 通过将所有这些技术整合在一起，组织正在重塑开展业务的方式。 这仅仅是个开始; 我们甚至不在第一局，甚至还没有记录过第一击！

到此为止，让我们开始看一下 AI 的一些当代应用。

# 数字个人助理和聊天机器人

不幸的是，对于某些呼叫中心来说，仍然太普遍以至于无法使用制作为的传统**交互式语音响应**（ **IVR** ）**系统** 称他们为耐心锻炼。 但是，我们在自然语言处理领域取得了巨大的进步：聊天机器人。 一些最受欢迎的示例是：

*   **Google 助手**：Google 助手于 2016 年启动，是目前可用的最高级聊天机器人之一。 可以在各种设备中找到它，例如电话，耳机，扬声器，洗衣机，电视和冰箱。 如今，大多数 Android 手机都包含 Google Assistant。 Google Home 和 Nest Home Hub 也支持 Google Assistant。
*   **Amazon Alexa** ：Alexa 是由 Amazon 开发和销售的虚拟助手。 它可以通过语音和执行命令（例如播放音乐，创建待办事项，设置警报，播放有声读物和回答基本问题）与用户互动。 它甚至可以按需告诉您一个笑话或故事。 Alexa 还可以用于控制兼容的智能设备。 开发人员可以通过安装技能来扩展 Alexa 的功能。 Alexa 技能是第三方供应商开发的其他功能。
*   **Apple Siri** ：Siri 可以接受用户语音命令和自然语言用户界面，以通过解析这些语音命令并将这些请求委派给一组 Internet 服务来回答问题，提出建议并执行操作。 该软件可以适应用户的个别语言使用，他们的搜索和偏好。 使用得越多，它就会学得越多，就会越好。
*   **Microsoft Cortana** ：Cortana 是由 Microsoft 设计和创建的另一种数字虚拟助手。 Cortana 可以设置提醒和警报，识别自然的语音命令，并使用信息回答问题。

所有这些助手都将允许您执行全部或至少大部分以下任务：

*   家里的控制设备
*   根据命令播放音乐和显示视频
*   设置计时器和提醒
*   预约
*   发送短信和电子邮件
*   拨打电话
*   公开申请
*   阅读通知
*   执行翻译
*   从电子商务网站订购

可能不支持但将变得越来越普遍的一些任务是：

*   登机
*   预订酒店
*   预订餐厅

所有这些平台还支持 3 个<sup>和</sup>第三方开发人员来开发自己的应用程序或 Amazon 称呼他们的“技能”。 因此，可能性是无限的。

现有 Alexa 技能的一些示例：

*   **MySomm** ：推荐某些肉配哪种葡萄酒
*   **调酒师**：提供有关如何制作酒精饮料的说明
*   **7 分钟锻炼**：将指导您进行艰难的 7 分钟锻炼
*   **Uber** ：允许通过 Alexa 订购 Uber 游乐设施

前面列出的所有服务都在不断改善。 他们不断从与客户的互动中学习。 服务的开发人员以及利用服务用户每天创建的新数据点的系统都对它们进行了改进。

大多数云提供商都非常容易创建聊天机器人，对于某些基本示例，无需使用编程语言。 此外，将这些聊天机器人部署到 Slack，Facebook Messenger，Skype 和 WhatsApp 等服务并不难。

# 个人司机

无人驾驶或无人驾驶汽车是无需人工协助即可沿着预先建立的路线行驶的车辆。 如今，大多数自动驾驶汽车都不再依赖单一传感器和导航方法，而是使用多种技术，例如雷达，声纳，激光雷达，计算机视觉和 GPS。

随着技术的出现，行业开始创建标准以实施和衡量其进度。 无人驾驶技术也是如此。 SAE International 已创建标准 J3016，该标准定义了汽车的六个自动化级别，以便汽车制造商，供应商和政策制定者可以使用相同的语言来对汽车的复杂程度进行分类：

**级别 0（无自动化）**

该车没有自动驾驶功能。 驾驶员充分参与并负责。 驾驶员操纵，制动，加速和协商交通。 这描述了当今道路上最新的汽车。

**1 级（驾驶员辅助）**

系统功能：在某些情况下，汽车可以控制转向或车辆速度，但不能同时控制两者。

驾驶员参与：驾驶员执行驾驶的所有其他方面，并全权负责监视道路并在辅助系统无法正常工作时接管汽车。 例如，自适应巡航控制。

**2 级（部分自动化）**

在某些情况下，汽车可以转向，加速和制动。 驾驶员仍然执行许多操作，例如解释和响应交通信号或改变车道。 控制车辆的责任主要落在驾驶员身上。 制造商仍然要求驾驶员充分参与。 此级别的示例包括：

*   奥迪交通堵塞辅助
*   凯迪拉克超级巡航
*   梅赛德斯-奔驰驾驶员辅助系统
*   特斯拉自动驾驶仪
*   沃尔沃飞行员辅助

**3 级（有条件的自动化）**

2 级和 3 级之间的枢轴点很关键。 在这个级别上，控制和监视汽车的责任开始从驾驶员变为计算机。 在适当的条件下，计算机可以控制汽车，包括监视环境。 如果汽车遇到无法处理的情况，它将要求驾驶员进行干预并采取控制措施。 驾驶员通常不控制汽车，但必须随时可以接管。 奥迪交通堵塞飞行员就是一个例子。

**级别 4（高度自动化）**

在大多数情况下，汽车不需要人为干预，但在某些道路，天气或地理条件下，仍需要人工协助。 在仅限于定义区域的共享汽车模型下，可能不会有人参与。 但是对于私人拥有的汽车，驾驶员可以管理地面街道上的所有驾驶职责，并且系统可以接管高速公路。 Google 现已淘汰的 Firefly 吊车就是这一级别的一个例子。 它没有踏板或方向盘。 它被限制为最高时速 25 英里/小时，并且没有在公共街道上使用。

**5 级（全自动）**

无人驾驶系统可以控制并在人类可以应付的任何道路上和任何条件下操作汽车。 汽车的“操作员”只需输入目的地。 目前尚无此级别的产品，但是有几家公司已经关闭，到本书出版时可能已经存在。

现在，我们将回顾在该领域工作的一些领先公司：

**Google 的 Waymo**

截至 2018 年，Waymo 的自动驾驶汽车在公共道路上行驶了 800 万英里，在模拟环境中行驶了 50 亿英里。 在接下来的几年中，几乎可以肯定，我们将能够购买能够完全自动驾驶的汽车。 特斯拉（Tesla）等人已经通过其自动驾驶功能提供了驾驶员协助，并且可能将成为第一家提供完整自动驾驶功能的公司。 想象一个世界，今天出生的孩子将永远不必获得驾照！ 仅凭人工智能的进步，对我们社会的破坏将是巨大的。 不再需要送货司机，出租车司机和卡车司机。 即使在无人驾驶的未来仍然发生交通事故，由于我们将消除分心驾驶和酒后驾驶，因此将挽救数百万生命。

Waymo 于 2018 年在美国亚利桑那州推出了首项商业无人驾驶服务，并计划在全国和全球范围内进行扩展。

**Uber ATG**

Uber 的**先进技术小组**（ **ATG** ）是 Uber 的子公司，致力于开发自动驾驶技术。 2016 年，Uber 在匹兹堡的街道上推出了一项实验性汽车服务。 优步计划购买多达 24,000 辆沃尔沃 XC90，并为其配备自动驾驶技术，并于 2021 年开始以一定容量实现商业化。

可悲的是，2018 年 3 月，伊莱恩·赫兹伯格（Elaine Herzberg）参与了一辆 Uber 无人驾驶汽车的事故并死亡。 根据警方的报道，她在试图通过手机观看视频时，试图过马路时被 Uber 车辆撞倒。 赫兹伯格女士成为涉及无人驾驶汽车的事故中首批死亡的人之一。 理想情况下，我们希望这种技术不会发生任何事故，但是我们需要根据当前交通事故带来的危机来调整我们要求的安全水平。 就背景而言，2017 年美国有 40,100 辆汽车死亡; 即使我们继续看到无人驾驶汽车发生事故，如果将死亡人数削减一半，每年也可以挽救数千人的生命。

当然，可以设想一种无人驾驶汽车，其外观比起我们现有汽车的内部更像一个客厅。 不需要方向盘，踏板或任何种类的手动控制。 汽车唯一需要的输入就是您的目的地，可以在旅行开始时通过对您的汽车“说话”来输入。 由于汽车将能够感知何时需要维修或汽车功能出现问题，因此无需跟踪维护时间表。

车祸的责任将从车辆的驾驶员转移到车辆的制造商，而不再需要购买汽车保险。 最后一点可能是汽车制造商部署该技术的速度缓慢的原因之一。 甚至汽车拥有量也可能被颠覆，因为我们可以在需要时而不是一直需要时召唤一辆车。

# 货运和仓库管理

亚马逊的分拣设施是在人类，计算机和机器人之间形成的共生关系的最好例子之一。 计算机接受客户订单并决定将商品路由到何处，机器人充当 mu 子，在仓库中搬运托盘和库存。 人们通过手动挑选进入每个订单的物品来解决“最后一英里”问题。 机器人熟练地无意识地重复执行多次任务，只要涉及到某种模式并且要进行一定程度的预训练就可以做到。 但是，让机器人捡起 20 磅重的包装，然后立即能够抓鸡蛋而不破鸡蛋，这是机器人技术难题之一。

机器人在处理不同大小，重量，形状和易碎性的物体时会遇到困难。 许多人可以轻松完成的任务。 因此，人们要处理机器人遇到的困难的任务。 这三种类型的不同演员之间的相互作用转化为一个经过微调的乐团，该乐团每天可以交付数百万个程序包而几乎没有错误。

甚至亚马逊机器人实现总监斯科特·安德森（Scott Anderson）在 2019 年 5 月都承认，距离全自动仓库至少有 10 年的路程。 因此，我们将继续在世界各地的仓库中看到这种配置一段时间。

# 人类健康

在健康科学中应用 AI 的方法几乎是无限的。 我们将在这里讨论其中的一些，但这绝不是详尽的清单。

**药物发现**

AI 可以协助生成候选药物（即要在医学上测试的分子），然后使用约束满足或实验模拟快速消除其中的一些候选药物。 在后面的章节中，我们将学习有关约束满足编程的更多信息。 简而言之，这种方法使我们能够通过快速生成数百万种可能的候选药物来加快药物发现，并且在候选药物不满足某些预定约束条件时也可以迅速拒绝它们。

此外，在某些情况下，我们可以在计算机中模拟实验，否则在现实生活中执行该实验会更加昂贵。

此外，在某些情况下，研究人员仍在进行真实世界的实验，但依靠机器人进行实验并加快处理过程。 这些新兴领域被称为**高通量筛选**（ **HTS** ）和**虚拟高通量筛选**（ **VHTS** ）。

机器学习已开始越来越多地用于增强临床试验。 埃森哲咨询公司开发了一种称为**智能临床试验**（ **ITP** ）的工具。 它用于预测临床试验的时间。

可以令人惊讶地使用的另一种方法是应用于药物发现的方法是**自然语言处理**（ **NLP** ）。 可以使用一串字母来表示基因组数据，并且可以使用 NLP 技术来处理或“理解”基因组序列的含义。

**保险定价**

机器学习算法可用于通过更准确地预测在患者身上花费多少，驾驶员的健康状况或寿命长短来更好地为保险定价。

例如，Insilico Medicine 的 *young.ai* 项目可以准确地预测血液样本和照片中某人的寿命。 血液样本提供 21 种生物标志物，例如胆固醇水平，炎症标志物，血红蛋白计数和白蛋白水平，用作机器学习模型的输入。 模型的其他输入是种族和年龄，以及该人的照片。

有趣的是，到目前为止，任何人都可以通过访问 [young.ai](https://young.ai) 并提供所需信息来免费使用此服务。

**患者诊断**

通过使用复杂的规则引擎和机器学习，医生可以对患者进行更好的诊断，并在实践中提高生产力。 例如，在张刚[1]在加利福尼亚大学圣地亚哥分校的一项最新研究中，一种系统可以比初级儿科医生更准确地诊断儿童疾病。 该系统能够以 90％至 97％的准确度诊断以下疾病：

*   腺热
*   罗索拉
*   流感
*   水痘
*   手足口病

输入数据集包含 2016 年至 2017 年间来自中国广州地区的 130 万儿童就医的医疗记录。

**医学影像学解释**

医学成像数据是有关患者的复杂而丰富的信息来源。 CAT 扫描，MRI 和 X 射线包含原本无法获得的信息。 放射科医生和临床医生不足以解释它们。 从这些图像获得结果有时可能需要几天时间，有时可能会被误解。 最近的研究发现，机器学习模型的性能甚至可以比人类模型好。

数据科学家开发了支持 AI 的平台，与传统方法相比，该平台可以在几分钟（而不是几天）内解释 MRI 扫描和放射图像，并且准确性更高。

令人惊讶的是，美国放射学院的领导者们对此并不感到担心，而是将 AI 的出现视为对医生有价值的工具。 为了促进该领域的进一步发展，**美国放射学院数据研究所**（ **ACR DSI** ）发布了医学成像中的多个 AI 用例，并计划继续发布 更多。

**精神病学分析**

与精神科医生进行一个小时的会议可能需要花费数百美元。 我们正处于能够使用 AI 聊天机器人模拟行为的风口浪尖。 至少，这些漫游器将能够在与精神科医生的会谈中提供后续护理，并在医生就诊之间为患者提供护理。

自动化顾问的一个早期例子是 Eliza。 它是由 Joseph Weizenbaum 在 1966 年开发的。 它允许用户与模仿罗杰式心理治疗师的计算机进行“对话”。 值得注意的是，Eliza 感觉很自然，但是它的代码只有几百行，并且其核心并没有真正使用太多的 AI。

最近和更先进的示例是 Ellie。 Ellie 由南加州大学创意技术研究所创建。 它有助于治疗抑郁症或创伤后应激障碍患者。 埃莉（Ellie）是位虚拟的治疗师（她出现在屏幕上），对情绪暗示做出反应，在适当的时候肯定地点头，并转移座位。 她可以感知一个人的脸上的 66 个点，并使用这些输入来读取一个人的情绪状态。 艾莉（Ellie）的秘密之一是，她显然不是人，这使人们对她的判断力降低了，对她的开放更自在。

**智能健康记录**

众所周知，医学落后于电子记录。 数据科学提供了多种方法来简化对患者数据的捕获，包括 OCR，手写识别，语音到文本捕获以及实时读取和分析患者生命体征。 不难想象不久的将来，由于迫在眉睫的健康问题，人工智能引擎可以实时分析这些信息，从而做出诸如调整人体葡萄糖水平，服用药物或寻求医疗帮助等决策。

**疾病检测与预测**

人类基因组是的最终数据集。 在不久的某个时候，我们将能够使用人类基因组作为机器学习模型的输入，并能够使用此庞大的数据集检测和预测各种各样的疾病和状况。

使用基因组数据集作为机器学习的输入是一个令人兴奋的领域，它正在迅速发展，并将彻底改变医学和卫生保健。

人类基因组包含超过 30 亿个碱基对。 我们在两个方面将取得进展，这将加快进度：

*   对基因组生物学的理解不断进步
*   大数据计算的进步，可以更快地处理大量数据

有很多研究将深度学习应用于基因组学领域。 尽管基因组学的深度学习仍处于早期阶段，但它有可能为以下领域提供信息：

*   功能基因组学
*   肿瘤科
*   人口遗传学
*   临床遗传学
*   作物单产提高
*   流行病学与公共卫生
*   进化和系统发育分析

# 知识搜索

我们到了的地步，在某些情况下，甚至都没有意识到我们正在使用人工智能。 一个技术或产品是好的标志是当我们不必停止思考它是如何做的时。 Google 搜索就是一个很好的例子。 该产品已经在我们的生活中无处不在，我们还没有意识到它在多大程度上依赖人工智能来产生惊人的结果。 从其谷歌建议技术到其结果相关性的不断提高，人工智能已深深地嵌入其搜索过程中。

据彭博社报道，2015 年初，Google 开始使用名为 RankBrain 的深度学习系统来协助生成搜索查询响应。 彭博社的文章对 RankBrain 的描述如下：

> “ RankBrain 使用人工智能将大量书面语言嵌入计算机可以理解的数学实体（称为矢量）中。如果 RankBrain 看到了自己不熟悉的单词或短语，则机器可以猜测是哪个单词或短语。 短语可能具有相似的含义并相应地过滤结果，从而使其更有效地处理了前所未有的搜索查询。”

> -克拉克，杰克[2]

在上一份报告中，Ranrainbrain 在数十亿个 Google 搜索查询中扮演着重要角色。 可以想像，该公司对 RankBrain 的工作原理 tight 之以鼻，而且甚至 Google 都可能很难解释它的工作方式。 您知道，这是深度学习的难题之一。 在许多情况下，它可以提供高度准确的结果，但是就为什么给出单个答案而言，深度学习算法通常很难理解。 基于规则的系统，甚至其他机器学习模型（例如 Random Forest）都更容易解释。

深度学习算法缺乏可解释性会产生重大影响，包括法律影响。 最近，Google 和 Facebook 等人发现自己处于显微镜下，以确定他们的结果是否有偏见。 将来，立法者和监管者可能会要求这些技术巨头为某种结果提供理由。 如果深度学习算法不能提供可解释性，则可能会迫使他们使用其他精度较低的算法来提供解释性。

最初，RankBrain 仅协助约 15％的 Google 查询，但现在几乎涉及所有用户查询。

但是，如果查询是普通查询或算法可以理解的查询，则 RankBrain 排名得分的权重很小。 如果查询是该算法之前未曾查看过的查询，或者它不知道其含义，则 RankBrain 得分更为相关。

# 推荐系统

建议系统是已融入我们日常生活的 AI 技术的另一个示例。 亚马逊，YouTube，Netflix，LinkedIn 和 Facebook 都依赖推荐技术，我们甚至没有意识到我们正在使用它。 推荐系统严重依赖数据，并且可用数据越多，它们将变得越强大。 这些公司拥有世界上最大的市值，这并非巧合，它们的力量来自能够利用客户数据中隐藏的力量。 希望这种趋势将来会继续。

什么是建议？ 让我们首先探索什么不是答案。 这不是一个明确的答案。 某些问题，例如“二加二是什么？” 或“土星有多少颗卫星？” 有明确的答案，没有主观性的余地。 其他问题，例如“您最喜欢的电影是什么？” 或“你喜欢萝卜吗？” 是完全主观的，答案将取决于回答问题的人。 一些机器学习算法因这种“模糊性”而蓬勃发展。 同样，这些建议可能会产生巨大的影响。

考虑一下亚马逊不断推荐某产品而不是另一产品的后果。 如果没有找到其他分销和销售产品的方法，那么制作推荐产品的公司将会蓬勃发展，制作不推荐产品的公司可能会倒闭。

推荐系统可以改进的方法之一是通过从系统用户那里进行先前的选择。 如果您是第一次访问电子商务网站，但是没有订单历史记录，那么该网站将很难为您量身定制建议。 如果您购买运动鞋，则该网站现在有一个数据点，可以作为起点使用。 根据系统的复杂程度，可能会建议使用另一双运动鞋，一双运动袜甚至篮球（如果鞋子是高帮鞋）。

良好推荐系统的重要组成部分是随机因素，该因素偶尔会“四肢走动”，并提出可能与初始用户选择无关的奇怪建议。 推荐系统不仅从历史中学习来找到相似的建议，而且还尝试提出可能乍一看并不相关的新建议。 例如，Netflix 用户可能会观看“教父”，而 Netflix 可能会开始推荐 Al Pacino 电影或流氓电影。 但它可能会建议您使用“伯恩身份”（Bourne Identity）。 如果用户不接受推荐或不观看电影，则算法将从中学习并避免使用其他电影，例如“伯恩身份”（例如，以杰森·伯恩为主要角色的任何电影）。

随着推荐系统变得越来越好，可能性是令人兴奋的。 他们将能够为个人数字助理提供支持，并成为您的私人管家，该管家对您的好恶有深刻的了解，并可以提出您可能没有想到的好建议。 建议可以从这些系统中受益的领域包括：

*   餐厅
*   电影
*   音乐
*   潜在合作伙伴（在线约会）
*   书籍和文章
*   搜索结果
*   金融服务（机器人顾问）

推荐系统的一些值得注意的特定示例如下：

**Netflix 奖**

Netflix 奖是在推荐器系统社区中引起大量嗡嗡声的竞赛。 从 2006 年到 2009 年，Netflix 赞助了比赛，奖金为一百万美元。 Netflix 提供了超过 1 亿个收视率的数据集。

Netflix 愿意向推荐最准确的团队支付奖金，并且比 Netflix 现有推荐系统的推荐准确性高 10％。 这项竞赛激发了人们对新的，更准确的算法的研究。 2009 年 9 月，大奖授予了 BellKor 的 Pragmatic Chaos 团队。

**潘朵拉**

潘朵拉是领先的音乐服务之一。 与 Apple 和 Amazon 等其他公司不同，Pandora 的唯一重点是音乐服务。 潘多拉（Pandora）显着的服务功能之一就是定制广播电台的概念。 这些“电台”允许用户按流派播放音乐。 您可以想象，推荐系统是此功能的核心。

Pandora 的推荐器基于多个层次：

*   首先，他们的音乐专家团队会根据流派，节奏和进度来注释歌曲。
*   这些注释被转换为用于比较歌曲相似度的向量。 这种方法促进了来自未知艺术家的“长尾巴”或晦涩音乐的呈现，尽管如此，它仍然可能适合个人听众。
*   该服务也高度依赖用户反馈，并使用它不断增强服务。 Pandora 收集了超过 750 亿个关于收听者偏好的反馈数据点。
*   然后，潘多拉（Pandora）推荐引擎可以使用收听者的先前选择，地理位置和其他人口统计数据，根据收听者的偏好执行个性化过滤。

总体而言，潘多拉（Pandora）的推荐器使用大约 70 种不同的算法，其中包括 10 种用于分析内容的算法，40 种用于处理集体智慧的算法以及大约 30 种用于进行个性化过滤的算法。

**改善**

机器人顾问是推荐引擎，可在最少人参与的情况下提供投资或财务建议和管理。 这些服务使用机器学习来自动分配，管理和优化客户的资产组合。 他们可以以比传统顾问更低的成本提供这些服务，因为它们的开销更低，并且其方法更具可扩展性。

现在，在这个领域竞争激烈，有超过 100 家公司提供此类服务。 机器人顾问被认为是一个巨大的突破。 以前，财富管理服务是专为高净值人士提供的专有且昂贵的服务。 与传统的人工服务相比，机器人顾问承诺以更低的成本向更广泛的受众提供类似的服务。 机器人顾问可以潜在地将投资分配到各种各样的投资产品中，例如股票，债券，期货，商品，房地产以及其他外来投资。 但是，为了简单起见，投资通常限于**交易所交易基金**（ **ETF** ）。

正如我们提到的，有许多公司提供机器人咨询。 例如，您可能需要研究 Betterment 以了解有关此主题的更多信息。 填写风险调查表后，Betterment 将为用户提供定制的，多样化的产品组合。 通常，Betterment 将推荐低收费股票和债券指数基金的组合。 Betterment 收取管理费（占产品组合的百分比），但低于大多数人力服务。 请注意，我们不认可这项服务，我们仅将其作为金融行业推荐引擎的示例。

# 智能家居

每当您在大街上向普通百姓提出 AI 的话题时，他们通常都会对替代人类工人的时间表示怀疑。 他们可以正确地指出我们仍然需要在房子周围做很多家务活的事实。 人工智能不仅需要在技术上成为可能，而且还需要在经济上可行才能被广泛采用。 家政服务通常是一种低薪职业，因此，取代它的自动化需要价格相同或更便宜。 此外，家务劳动需要很多技巧，而且其中包括不一定要重复的任务。 让我们列出该自动机需要熟练执行的一些任务：

*   洗衣服
*   叠衣服
*   煮晚餐
*   铺床
*   捡起地板上的物品
*   拖把，灰尘和真空
*   洗碗
*   监控房屋

众所周知，其中一些任务很容易在机器上执行（即使没有 AI），而其中一些则非常困难。 因此，出于经济考虑，房屋可能成为最后一个完全自动化的地方之一。 尽管如此，让我们看一下在这一领域取得的一些惊人进步。

**家庭监控**

家庭监视是领域，通常已经可以找到很好的解决方案。 亚马逊提供的 Ring 视频门铃和 Google Nest 温控器是两种价格低廉的选择，已广泛使用和流行。 这是今天可以购买的智能家居设备的两个简单示例。

环形视频门铃是连接到互联网的智能家居设备，可以通过智能手机通知房主家中的活动，例如访客。 该系统不会连续记录，而是在按下门铃或激活移动探测器时激活。 然后，环形门铃可以让房主观看活动或使用内置的麦克风和扬声器与访客交流。 有些型号还允许房主通过智能锁远程打开门，并让访客进入房屋。

Nest Learning Thermostat 是一款智能家用设备，最初由 Nest Labs 开发，该公司后来被 Google 收购。 它是由 Tony Fadell，Ben Filson 和 Fred Bould 设计的。 它是可编程的，支持 Wi-Fi 且具有自学习功能。 它使用人工智能来优化房屋的温度，同时节省能源。

在使用的最初几周中，将恒温器设置为首选设置，这将作为基准。 温控器将了解您的日程安排和您的首选温度。 使用内置传感器和手机的位置，当没人在家时，恒温器将转换为节能模式。

自 2011 年以来，Nest 温控器已为全球数百万个家庭节省了数十亿千瓦时的能源。 独立研究表明，它可以为人们节省平均 10％至 12％的取暖费用和 15％的冷却费用，因此大约 2 年内它可以收回成本。

**吸尘和拖把**

吸尘和拖把是流行的移交给机器人的两个任务。 机器人吸尘器是使用 AI 吸尘表面的自主机器人吸尘器。 根据设计的不同，其中一些机器使用旋转刷达到狭窄的角落，某些机型除了具有吸尘功能外还包括其他一些功能，例如拖把和紫外线杀菌。 推广这项技术的大部分功劳归于公司（而非电影） *iRobot* 。

iRobot 于 1990 年由 Rodney Brooks，Colin Angle 和 Helen Greiner 在麻省理工学院的人工智能实验室工作后相识而创立。 iRobot 以其吸尘机器人（Roomba）而闻名，但很长一段时间以来，他们还拥有一个专门从事军事机器人开发的部门。 Roomba 于 2002 年开始销售。截至 2012 年，iRobot 已售出超过 800 万台家用机器人，并制造了 5,000 多台国防和安全机器人。 该公司的 PackBot 是美军使用的炸弹处理机器人，已在伊拉克和阿富汗广泛使用。 PackBots 还被用于在福岛第一核电站的危险条件下收集信息。 在墨西哥湾的 Deepwater Horizo​​n 溢油事故发生之后，iRobot 的 Seaglider 被用来探测水下的油藏。

另一款 iRobot 产品是 Braava 系列清洁剂。 Braava 是一款小型机器人，可以拖地和扫地。 它适用于浴室和厨房等狭小空间。 它喷水并使用各种不同的垫来有效，安静地清洁。 某些 Braava 型号具有内置的导航系统。 Braava 没有足够的能力去除深层的污渍，因此它不是完整的人类替代品，但确实具有广泛的接受度和很高的评价。 我们希望他们继续受欢迎。

家庭智能设备的潜在市场是巨大的，几乎可以肯定的是，我们将继续看到成熟的公司和初创公司都试图利用这个尚未开发的市场。

**收拾烂摊子**

正如我们在运输用途案例中了解到的一样，挑选具有不同重量，尺寸和形状的物体是最困难的自动化任务之一。 机器人可以在均匀的条件下（例如工厂车间，某些机器人专门从事某些任务）高效地执行任务。 然而，在拾起椅子之后拾起一双鞋可能是巨大的挑战，而且代价昂贵。 因此，不要指望机器很快就会以具有成本效益的方式普及这种家庭杂务。

**私人厨师**

就像从地板上捡起物品一样，烹饪需要捡起不同的物品。 然而，有两个原因可以使我们期望“自动烹饪”会更快发生：

*   某些餐馆可能会收取数百美元的食物费用，并为熟练的厨师付出高昂的代价。 因此，如果这样做可以提高利润，他们可能会愿意使用技术来替换高价员工。 一个五星级的寿司餐厅就是一个例子。
*   厨房中的某些任务是重复性的，因此适合自动化。 想想快餐店，汉堡和炸薯条可能要成百上千。 因此，不是由一个机器来处理整个不同的烹饪过程，而是一系列机器可以处理该过程的各个重复阶段。

智能假肢是人工智能增强人类而不是替代人类的典范。 有不止几名厨师在事故中失去了手臂或出生时没有肢体。

一个例子是厨师迈克尔凯恩斯（Michael Caines），他经营一家米其林两星级餐厅，并在一场可怕的车祸中失去了手臂。 凯恩斯厨师一直担任英格兰德文郡吉德利公园的主厨，直到 2016 年 1 月。[3] 他目前是埃克塞特和埃克斯茅斯之间的林普斯通庄园酒店的行政总厨。 他现在用假肢做饭，但考虑到他的食物质量，您永远不会知道。

另一个例子是运动员和厨师 Eduardo Garcia –世界上最先进的仿生手使这一切成为可能。

2011 年 10 月，猎弓麋鹿在蒙大拿州偏僻地区被电死。 爱德华多（Eduardo）于 2011 年 10 月独自打猎。当他看到一只死去的小黑熊时，他正在乡下。 他停下来检查一下，跪下来，用刀刺了一下。

这样做时，他的身体承受了 2400 伏的电压-熊宝宝被一根埋在地下的带电电线杀死。 他幸免于难，但在事件中失去了手臂。

2013 年 9 月，Advanced Arm Dynamics 为 Garcia 配备了 Touch Bionics 设计的仿生手。 仿生手由加西亚的前臂肌肉控制，可以用 25 种不同的方式抓握。 用他的新手，加西亚可以执行通常需要非常敏捷的任务。 他的新手仍然有一些局限性。 例如，加西亚（Garcia）无法举重。 但是，他现在可以执行某些以前无法执行的操作。 例如，他可以从烤箱中拿出东西，而不会被灼伤，并且不可能割伤手指。

相反，机器人可以代替厨房中的人类来代替人类，而不是。 一个例子就是机器人厨房 Moley。 Moley 目前尚未投入生产，但 Moley 机器人厨房的最先进原型由两个机械臂组成，这些机械臂的手配有触觉传感器，炉灶，烤箱，洗碗机和触摸屏单元。 这些人造手可以举起，抓住并与大多数厨房设备互动，包括刀具，打蛋器，汤匙和搅拌器。

使用 3D 相机和手套，它可以记录人类厨师做饭的情况，然后将详细的步骤和说明上载到存储库中。 然后使用手势识别模型将厨师的动作转换为机器人动作。 这些模型是与斯坦福大学和卡内基梅隆大学合作创建的。 之后，Moley 可以重复相同的步骤，并从头开始烹饪完全相同的餐点。

在当前的原型中，用户可以使用触摸屏或智能手机应用程序对其进行操作，并预先准备好配料并将其放置在预设位置。 该公司的长期目标是允许用户从 2000 多种食谱中轻松选择一个选项，Moley 会在几分钟内准备好餐点。

# 游戏

可能没有例子比在游戏领域取得的进步更好的例子来证明人工智能的惊人发展。 人类具有天生的竞争能力，在我们自己的游戏中让机器击败我们是衡量该领域突破的有趣标准。 长期以来，计算机一直能够在一些更基础，更确定性，计算量更少的游戏（例如说跳棋）中击败我们。 直到最近几年，机器才能够始终击败某些较难游戏的主人。 在本节中，我们将介绍其中的三个示例。

**星际争霸 2**

电子游戏被用作测试 AI 系统性能的基准已有数十年。 随着功能的增强，研究人员可以使用需要不同类型智能的更为复杂的游戏。 从此游戏中开发出来的策略和技术可以转移到解决实际问题上。 《星际争霸 2》的游戏被认为是最难的游戏之一，尽管按照视频游戏标准它是古老的游戏。

DeepMind 的团队引入了一个名为 AlphaStar 的程序，该程序可以玩《星际争霸 II》，并且首次击败了顶级职业玩家。 在 2018 年 12 月举行的比赛中，AlphaStar 击败了由 Grzegorz“ MaNa” Komincz 组建的团队，Grzegorz“ MaNa” Komincz 是全球最强的星际争霸职业球员之一，得分为 5-0。 游戏是在专业比赛条件下进行的，没有任何游戏限制。

与以前使用 AI 限制游戏的尝试相反，AlphaStar 可以不受限制地玩完整游戏。 它使用了深度神经网络，可以通过监督学习和强化学习直接从原始游戏数据中进行训练。

使《星际争霸 2》如此困难的一件事是需要平衡短期和长期目标并适应意外情况。 这通常对先前的系统提出了巨大的挑战。

尽管 StarCraft 只是一个游戏，尽管很困难，但是 AlphaStar 提出的概念和技术对于解决其他现实世界中的挑战很有用。 例如，AlphaStar 的体系结构能够根据不完善的信息来对很长的可能动作序列进行建模-游戏通常持续数小时，并进行成千上万次动作。 在许多实际问题中，可以找到对长序列进行复杂预测的主要概念，例如：

*   天气预报
*   气候模拟
*   自然语言理解

AlphaStar 在星际争霸中表现出的成功代表着现有最困难的视频游戏之一的重大科学突破。 这些突破代表了人工智能系统的巨大飞跃，该系统可以转让并可以帮助解决基本的现实世界中的实际问题。

**危险**

IBM 和 Watson 团队在 2011 年创造了历史，当时他们设计了系统，该系统能够击败两个最成功的 Jeopardy 冠军。

肯·詹宁斯（Ken Jennings）连续 74 次出场，是该演出历史上最长的不败战绩。 布拉德·拉特（Brad Rutter）赢得了最大的奖金，奖金总额为 325 万美元。

两位球员都同意与沃森进行一场比赛。

Watson 是一个问答系统，可以回答以自然语言提出的问题。 它最初由主要研究人员 David Ferrucci 领导的 IBM DeepQA 研究团队创建。

Watson 使用的问题解答技术与常规搜索（例如 Google 搜索）之间的主要区别在于，常规搜索将关键字作为输入，并根据与查询的相关性对文档列表进行排名。 像 Watson 所使用的一样，问答技术采用自然语言表达一个问题，试图更深入地理解该问题，并试图为该问题提供准确的答案。

Watson 的软件架构使用：

*   IBM 的 DeepQA 软件
*   Apache UIMA（非结构化信息管理架构）
*   多种语言，包括 Java，C ++和 Prolog
*   SUSE Linux Enterprise Server
*   适用于分布式计算的 Apache Hadoop

**国际象棋**

我们许多人都记得 1996 年 Deep Blue 击败国际象棋大师 Gary Kasparov 时的新闻。Deep Blue 是 IBM 创建的国际象棋应用程序。

在第一轮比赛中，深蓝赢得了对阵加里·卡斯帕罗夫的第一场比赛。 但是，他们原定要打六场比赛。 卡斯帕罗夫（Kasparov）在接下来的五场比赛中赢得了三场胜利，并赢得了其中的两场，从而以 4–2 的比分击败了深蓝。

Deep Blue 团队回到进行了制图，对该软件进行了很多增强，并在 1997 年再次使用了 Kasparov。Deep Blue 在第二轮击败 Kasparov 的比赛中赢得了第二轮比赛，以 3½的比分赢得了六局比赛 2½。 然后，它成为第一个在标准国际象棋锦标赛规则和时间控制下击败比赛的世界冠军的计算机系统。

一个鲜为人知的例子是 AlphaZero 团队在国际象棋领域的成就，这是一种机器击败人类的迹象。

Google 的 AlphaZero 研究团队的科学家在 2017 年创建了一个系统，该系统只花了四个小时就学会了象棋规则，然后粉碎了当时最先进的世界冠军象棋程序 *Stockfish* 。 到现在为止，有关计算机还是人类在国际象棋方面更胜一筹的问题已得到解决。

让我们暂停片刻，然后考虑一下。 该系统超越了人类对古代象棋的所有知识，如果该系统在早晨开始学习，那么它将在午餐时间完成。

该系统已获得国际象棋规则，但未获得任何策略或进一步的知识。 然后，在几个小时内，AlphaZero 达到了足以击败 Stockfish 的程度。

在与 Stockfish 的一系列 100 场比赛中，AlphaZero 以白色打赢了 25 场比赛（白色具有优势，因为它排名第一）。 它还赢得了三场黑色比赛。 其余的比赛是平局。 Stockfish 没有获得任何胜利。

**AlphaGo**

象象棋一样困难，的难度与古代 Go 棋相比没有。

不仅（19 x 19）棋盘位置比可见宇宙中的原子还多，而且可能的国际象棋棋盘位置与棋盘位置的数量可以忽略不计。 但是围棋至少比象棋游戏复杂几个数量级，因为围棋的发展有很多可能的方式，使每种棋子都朝着另一条发展路线发展。 使用围棋，单块石头可以影响和影响整个棋盘状况的移动次数也比带棋的单块移动大很多数量级。

有一个强大的程序示例，它可以玩由 DeepMind 开发的名为 AlphaGo 的 Go 游戏。 AlphaGo 还有三个功能更强大的继任者，分别称为 AlphaGo Master，AlphaGo Zero 和 AlphaZero。

2015 年 10 月，原始的 AlphaGo 成为第一个在全尺寸 19 x 19 板上击败无障碍职业人类围棋选手的计算机围棋程序。 2016 年 3 月，它在五场比赛中击败了李·塞多尔。 这成为围棋程序第一次击败没有障碍的 9 杆专业运动员。 尽管 AlphaGo 在第四局中输给 Lee Sedol，但 Lee 在最后一场比赛中辞职，最终比分是 4 比 1。

在 2017 年未来围棋峰会上，AlphaGo 的继任者 AlphaGo Master 在三场比赛中击败了大师 Kejie。 柯洁当时被评为世界排名第一的球员。 此后，AlphaGo 被中国围棋协会授予专业 9 丹。

AlphaGo 及其继任者使用蒙特卡洛树搜索算法，基于先前通过机器学习“学习”的知识来找到他们的动作，特别是使用深度学习和训练，既可以与人类玩耍，也可以与人类玩耍。 该模型经过训练可以预测 AlphaGo 自己的举动以及获胜者的游戏。 这种神经网络提高了树形搜索的强度，从而在后续游戏中提供了更好的动作和更强的玩法。

# 电影制作

几乎可以肯定的是在接下来的几十年内可以制作出 100％计算机生成的电影。 构想一个系统，其中输入是一个书面剧本，而输出是一个完整的故事片，这是不容置疑的。 此外，天然发电机已经取得了一些进步。 因此，最终甚至不需要脚本。 让我们进一步探讨。

**仿冒品**

Deepfake 是*术语*或“深度学习”和“ fake”的混合。 这是一种合并视频图像的 AI 技术。 常见的应用是将某人的脸重叠到另一张上。 它的一个邪恶版本被用来与著名人物合并色情场景或创建复仇色情片。 Deepfake 还可以用于创建虚假新闻或恶作剧。 可以想象，如果滥用该技术，将会对社会产生严重影响。

中国一家名为 Momo 的公司开发了类似软件的最新版本，后者开发了一个名为 *Zao* 的应用程序。 它使您可以将某人的脸重叠在诸如 Titanic 之类的短片上，效果令人印象深刻。 此应用程序和其他类似应用程序毫无争议。 隐私团体抱怨说，按照用户协议条款提交给网站的照片将成为 Momo 的财产，然后可用于其他应用程序。

有趣的是，技术在该领域将如何继续发展。

**电影脚本生成**

他们不会很快获得的奥斯卡金像奖，但是有两个项目致力于制作电影剧本。 最著名的例子之一是 Sunspring。

《 Sunspring》是一部实验科幻短片，于 2016 年发行。它完全是使用深度学习技术编写的。 这部电影的剧本是使用名为本杰明（Benjamin）的**长短期记忆**（ **LSTM** ）模型创建的。 它的创作者是由 BAFTA 提名的电影人奥斯卡·夏普（Oscar Sharp）和纽约大学 AI 研究人员罗斯·古德温（Ross Goodwin）。 电影中的演员是托马斯·米德迪奇，伊丽莎白·格雷和汉弗莱·克尔。 它们的字符名称分别是 H，H2 和 C，它们将保留在未来。 他们最终相互联系，形成了三角恋。

最初在科幻-伦敦电影节的 48 小时挑战赛上放映，并于 2016 年 6 月由技术新闻网站 Ars Technica 在线发布。

# 承销和交易分析

什么是承保？ 简而言之，承保是过程，机构通过该过程确定机构是否要承担财务风险以换取保费。 需要承保的交易示例如下：

*   签发保险单
    *   健康
    *   生活
    *   家
    *   驾驶
*   贷款额
    *   分期贷款
    *   信用卡
    *   按揭
    *   商业信贷
*   证券承销和首次公开募股（IPO）

可以预见，确定是否应该签发保险单或贷款，如果做出错误的决定，以什么价格会付出很高的代价。 例如，如果一家银行发行了一笔贷款，但贷款违约，则将需要数十笔其他表现良好的贷款来弥补这一损失。 相反，如果银行在借款人将要偿还全部贷款的情况下放弃贷款，这也将损害银行的财务状况。 因此，银行花费大量时间分析或“承销”贷款，以确定借款人的信用价值以及担保该贷款的抵押品的价值。

即使进行了所有这些检查，承销商仍然会出错并发行拖欠或绕过应得的借款人的贷款。 当前的承保过程遵循必须满足的一组标准，但是特别是对于较小的银行，在此过程中仍然存在一定程度的人为主观性。 这不一定是一件坏事。 让我们来看一个场景，以进一步探索它：

*最近，一位高净值人士从环球旅行中回来。 三个月前，他们在一家著名的医疗机构找到了一份工作，其信用评分超过 800。*

你会借钱给这个人吗？ 有了这些特征，它们似乎是一个很好的信用风险。 但是，正常的承保规则可能会使他们失去资格，因为最近两年没有使用这些规则。 手动承保会审视整个情况，并可能会批准它们。

同样，机器学习模型可能能够将其标记为有价值的帐户并发放贷款。 机器学习模型没有严格的规则，而是“通过示例学习”。

许多贷方已经在其承销中使用机器学习。 Zest Finance 是一家专门从事这一领域的公司的有趣示例。 Zest Finance 使用 AI 技术来协助贷方进行承销。 人工智能可以帮助增加收入并降低风险。 最重要的是，一般而言应用良好的 AI，尤其是 Zest Finance 可以帮助公司确保所使用的 AI 模型符合某个国家/地区的法规。 一些 AI 模型可能是一个“黑匣子”，很难解释为什么一个借款人被拒绝而另一个借款人被接受。 Zest Finance 可以全面解释数据建模结果，衡量业务影响并遵守法规要求。 Zest Finance 的秘密武器之一是使用非传统数据，包括贷方可能拥有的内部数据，例如：

*   客户支持数据
*   付款历史
*   购买交易

他们还可能考虑非传统信用变量，例如：

*   客户填写表格的方式
*   客户用来到达站点的方法或他们导航站点的方式
*   填写申请所需的时间

# 数据清理和转换

正如汽油为汽车提供动力一样，数据是 AI 的命脉。 古老的格言“垃圾进，垃圾出”仍然令人痛苦。 因此，拥有干净准确的数据对于生成一致，可重现和准确的 AI 模型至关重要。 其中一些数据清理需要艰苦的人类参与。 通过某种方法，据说数据科学家花费了大约 80％的时间来清理，准备和转换输入数据，而花费了 20％的时间运行并优化模型。 例如 ImageNet 和 MS-COCO 图像数据集。 两者都包含超过一百万个带有标签的各种对象和类别的图像。 这些数据集用于训练模型，该模型可以区分不同类别和对象类型。 最初，这些数据集是由人类艰苦而耐心地标记的。 随着这些系统的普及，我们可以使用 AI 进行标记。 此外，还有许多支持 AI 的工具可帮助进行清理和重复数据删除过程。

亚马逊湖形成就是一个很好的例子。 在 2019 年 8 月，亚马逊正式提供了 Lake Lake 的服务。 Amazon Lake Formation 自动执行创建数据湖通常涉及的一些步骤，包括数据的收集，清理，重复数据删除，编目和发布。 然后可以使数据可用于分析和构建机器模型。 要使用“湖泊形成”，用户可以使用预定义的模板将数据从一系列来源中引入湖泊。 然后，他们可以根据组织中各组所需的访问级别来定义用于控制数据访问的策略。

数据经过一些自动的准备，清理和分类，它们使用机器学习来自动执行这些任务。

Lake Formation 还提供了集中式仪表板，管理员可以在其中管理和监视跨多个分析引擎的数据访问策略，治理和审计。 用户还可以在结果目录中搜索数据集。 随着该工具在未来几个月和几年中的发展，它将使用他们最喜欢的分析和机器学习服务促进数据分析，其中包括：

*   数据块
*   板
*   亚马逊 Redshift
*   亚马逊雅典娜
*   AWS 胶水
*   亚马逊电子病历
*   亚马逊 QuickSight
*   亚马逊 SageMaker

# 摘要

本章提供了一些 AI 应用的示例。 也就是说，这里的内容并没有开始刮擦表面！ 我们试图将用例保留在广泛使用的技术上，或者至少将在不久之后变得可用的技术上。 不难推断该技术将如何继续改进，变得更便宜并得到更广泛的应用。 例如，当自动驾驶汽车开始流行时，这将非常令人兴奋。

但是，我们可以肯定的是，甚至还没有构思出 AI 的更大应用。 而且，人工智能的进步将对我们的社会产生广泛的影响，并且在某些时候，我们将不得不处理以下问题：

*   如果 AI 如此进化以至于变得有意识，会发生什么？ 应该给它权利吗？
*   如果用机器人代替人类，是否应该要求公司继续为该流离失所者支付工资税？
*   我们是否可以做到计算机无所不能，如果这样做，我们将如何适应？ 我们将如何度过时间？
*   更糟糕的是，该技术是否使少数人能够控制所有资源？ 会否出现一个个人可以追求自己的利益的普遍收入社会？ 还是流离失所的人民生活在贫困中？

比尔·盖茨和埃隆·马斯克警告说，人工智能要么疯狂地追求自己的目标就摧毁地球，要么是偶然地（或偶然地）破坏了人类。 我们将对 AI 的影响持更为乐观的“半满”看法，但是可以肯定的是，这将是一段有趣的旅程。

# 参考

1.  `Willingham, Emily, A Machine Gets High Marks for Diagnosing Sick Children, Scientific American, October 7th, 2019, https://www.scientificamerican.com/article/a-machine-gets-high-marks-for-diagnosing-sick-children/`
2.  `Clark, Jack, Google Turning Its Lucrative Web Search Over to AI Machines, Bloomberg, October 26th, 2015, https://www.bloomberg.com/news/articles/2015-10-26/google-turning-its-lucrative-web-search-over-to-ai-machines`
1.  <https://www.michaelcaines.com/michael-caines/about-michael/>