# 零、前言

TensorFlow 是一个开源软件库，用于实现机器学习和深度学习系统。

这两个名称的后面隐藏着一系列强大的算法，这些算法面临一个共同的挑战：使计算机学习如何自动识别复杂的模式并做出最明智的决策。

机器学习算法是有监督的还是无监督的； 尽可能简化，我们可以说最大的不同是在监督学习中，程序员指示计算机如何做某事，而在无监督学习中，计算机将自己学习所有。

相反，深度学习是机器学习研究的一个新领域，其目的是使机器学习更接近人工智能目标。 这意味着深度学习算法试图像人脑一样运作。

为了在这些引人入胜的领域进行研究，Google 团队开发了 TensorFlow，这是本书的主题。

为了介绍 TensorFlow 的编程功能，我们使用了 Python 编程语言。 Python 有趣且易于使用。 它是一种真正的通用语言，并且正在迅速成为任何自重程序员的必备工具。

本书的目的不是完整地描述所有 TensorFlow 对象和方法。 取而代之的是，我们将介绍重要的系统概念，并引导您尽快高效地学习。 本书的每一章都介绍了 TensorFlow 的不同方面，并附带了一些反映机器和深度学习的典型问题的编程示例。

尽管 TensorFlow 既庞大又复杂，但一旦您了解其基本设计和编程方法，它的设计便易于使用。

《TensorFlow 入门》的目的是帮助您做到这一点。

享受阅读！

# 本书涵盖的内容

第 1 章，“TensorFlow 基本概念”，包含有关 TensorFlow 的结构及其开发问题的一般信息。 它还提供了 Python 语言的基本编程准则以及安装过程之后的第一个 TensorFlow 工作会话。 本章最后对 TensorBoard 进行了描述，TensorBoard 是用于优化和调试的强大工具。

第 2 章，“使用 TensorFlow 进行数学运算”，描述了 TensorFlow 的数学处理能力。 它涵盖了基本代数的编程示例，直至偏微分方程。 此外，还解释了 TensorFlow 中的基本数据结构，即张量。

第 3 章，“机器学习入门”，介绍了一些机器学习模型。 我们开始实现线性回归算法，该算法与数据之间的建模关系有关。 本章的主要重点是解决机器学习中的两个基本问题。 分类，即如何将每个新输入分配给可能的给定类别之一； 数据聚类，这是将一组对象进行分组的任务，以使同一组中的对象比其他组中的对象更相似。

第 4 章，“神经网络介绍”提供了神经网络的快速详细介绍。 这些是代表元件之间的互连的数学模型，即人工神经元。 它们是在某种程度上模仿活神经元特性的数学结构。 神经网络为深度学习算法的架构奠定了基础。 然后实现了两种基本类型的神经网络：用于分类问题的单层感知机和多层感知机。

第 5 章，“深度学习”概述了深度学习算法。 直到最近几年，深度学习才收集了几年前难以想象的大量结果。 我们将展示如何实现两种基本的深度学习架构，即卷积神经网络（CNN）和循环神经网络（RNN），分别用于图像识别和语音翻译问题。

第 6 章，“GPU 编程和使用 TensorFlow”，展示了用于 *GPU* 计算的 TensorFlow 工具，并介绍了 *TensorFlow 服务*，一种针对机器学习模型的高性能开源服务系统，该模型针对生产环境而设计，并针对 TensorFlow 进行了优化。

# 这本书需要什么

所有示例均已在 Ubuntu Linux 64 位计算机上使用 Python 版本 2.7 实现，包括 TensorFlow 库版本 0.7.1。

您还将需要以下 Python 模块（最好是最新版本）：

*   点子
*   Bazel
*   Matplotlib
*   NumPy
*   Pandas

# 这本书是给谁的

读者应该具有编程和数学概念的基础知识，并且同时希望向您介绍机器和深度学习的主题。 阅读本书后，您将能够掌握 TensorFlow 的功能以构建功能强大的应用。

# 约定

在本书中，您将找到许多可以区分不同类型信息的文本样式。 以下是这些样式的一些示例，并对其含义进行了解释。

文本，数据库表名称，文件夹名称，文件名，文件扩展名，路径名称，虚拟 URL，用户输入和 Twitter 句柄中的代码字如下所示：“用于流控制的指令为`if`，`for`和 `while`。”

任何命令行输入或输出的编写方式如下：

```py
>>> myvar = 3
>>> myvar += 2
>>> myvar
5
>>> myvar -= 1
>>> myvar
4

```

**新术语**和**重要词**以粗体显示。 您在屏幕上看到的字词，例如在菜单或对话框中的字样如下所示：“本书中的快捷方式基于`Mac OSX 10.5+`方案。”

### 注意

警告或重要提示会出现在这样的框中。

### 小费

提示和技巧如下所示。